"""
Resources relating handling of responses from HTTP requests.

Includes operations to:
 * validate the success of HTTP requests
 * handle unsuccessful HTTP requests
 * deserialize response payload data to Python objects
"""
